/***************************************************************************
                          level.cpp  -  description
                             -------------------
    begin                : Mon May 21 2001
    copyright            : (C) 2001, 2002 by Thomas Friedrichsmeier
    email                : Thomas.Friedrichsmeier@ruhr-uni-bochum.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "defines.h"

#include <kglobal.h>
#include <klocale.h>
#include <kstandarddirs.h>

#include <qpoint.h>
#include <qdom.h>
#include <qpen.h>
#include <qlist.h>

#include "level.h"

#include "taxipilot.h"
#include "taxipilotapp.h"
#include "cdp.h"
#include "movingobject.h"
#include "teleporter.h"
#include "centerofgravity.h"
#include "platform.h"
#include "state.h"
#include "event.h"
#include "timer.h"
#include "hotarea.h"

Level::Level(const QString & file):QCanvasSprite(0, TPBase::canvas)
{
	level = this;					// other objects will access the common level-pointer, before the level is completely constructed!

	setAnimated(false);			// level never moves

	read_level(file);

	if (cdp->for_real()) {
		TPBase::canvas->setBackgroundPixmap(*background);
		TPBase::canvas->resize(width(), height());
		tp->app->setCaption(caption);

		config_changed ();
		setZ(LEVEL_Z);
		show();
		for (unsigned int i=0; i < num_timers; i++) {
			if (timers[i]->trigger_initially) {
				timers[i]->trigger ();
			}
		}
	}
}

Level::~Level()
{
	unsigned int i;
	if (num_platforms) {
		for (i = 0; i < num_platforms; i++) {
			delete platform[i];
		}
		delete[]platform;
	}

	if (num_teleporters) {
		for (i = 0; i < num_teleporters; i++) {		
			delete teleporters[i];			
		}
		delete[]teleporters;
	}

	if (num_grav_centers) {
		for (i = 0; i < num_grav_centers; i++) {		
			delete grav_center[i];			
		}
		delete[]grav_center;
	}

	if (num_objects) {
		for (i = 0; i < num_objects; i++) {
			delete objects[i];
		}
		delete[]objects;
	}

	if (cdp->for_real () || cdp->thorough_check ()) {
		hide ();
		delete image;
		TPBase::canvas->setBackgroundColor (QColor (0, 0, 0));
		delete background;
	}	

	if (num_states) {
		for (i = 0; i < num_states; i++) {
			delete states[i];
		}
		delete[]states;
	}

	if (num_events) {
		for (i = 0; i < num_events; i++) {
			delete events[i];
		}
		delete[]events;
	}

	if (num_timers) {
		for (i = 0; i < num_timers; i++) {
			delete timers[i];
		}
		delete[]timers;
	}

	if (num_areas) {
		for (i = 0; i < num_areas; i++) {
			delete areas[i];
		}
		delete[]areas;
	}
}

/** I consider this function to be mostly straight-forward. I do however know close to nothing about XML
(definitely no more than was neccessary to write up this code. Go ahead and change it to make more sense! */
void Level::read_level(const QString & file)
{

	QString convert;			// used to convert numbers
	QDomElement e;				// just some element
	QDomNodeList list;			// just a list of nodes

	QList <Promise> promises;
	QList <Promise> definitions;
	promises.setAutoDelete (true);
	definitions.setAutoDelete (true);
	num_areas = num_events = num_timers = num_grav_centers = num_objects = num_platforms = num_states = num_teleporters = 0;

	cdp->init_debug(file);

	QDomElement xdocElement = cdp->open_config_file(file);
	caption = cdp->get_string_attribute("caption", xdocElement, i18n("No Title"), Warn);
	author = cdp->get_string_attribute("author", xdocElement, i18n("Anonymous"), Info);

	if (! (cdp->for_real() || cdp->thorough_check())) { // all the rest not needed for credits
		return;
	}

	e = cdp->get_element("completion", xdocElement, Serious);	// get completion tag
	deliveries_to_go = cdp->get_int_attribute("deliveries", e, 1, 9999, 3, Serious);
	extra_life = cdp->get_bool_attribute("extra_life", e, false, Warn);

	e = cdp->get_element("image", xdocElement, Critical);	// get image-section
	background = cdp->createQCPixmap ("background", e, 640, 480, Serious);
	int i = 1;
	image = cdp->createQCPixmapArray("level", e, &i, 640, 480, Critical);

// BEGIN Read states, events and areas here!

// get state-declarations
	e = cdp->get_element ("states", xdocElement, Info);	// get all the states

	if (!(e.isNull())) {		// found a states-section?
		list = cdp->get_node_list ("state", e, 1, -1, 1, Warn);

		if (list.length ()) {
			states = new State * [list.length()];

			while (num_states < list.length ()) {
				e = list.item (num_states).toElement ();

				states[num_states] = new State (&e);		
				state_map.insert (states[num_states]->ident, states[num_states]);
	
	        		num_states++;
			}				
		}
	}

// read events
	e = cdp->get_element ("events", xdocElement, Info);	// get all the events

	if (!(e.isNull())) {		// found an events-section?
		list = cdp->get_node_list ("event", e, 1, -1, 1, Warn);

		if (list.length ()) {
			events = new Event * [list.length()];

			while (num_events < list.length ()) {
				e = list.item(num_events).toElement ();

				events[num_events] = new Event (&e);		
				event_map.insert (events[num_events]->ident, events[num_events]);
	
	        		num_events++;
			}
		}
	}

// read timers
	e = cdp->get_element ("timers", xdocElement, Info);	// get all the timers

	if (!(e.isNull())) {		// found an timers-section?
		list = cdp->get_node_list ("timers", e, 1, -1, 1, Warn);

		if (list.length ()) {
			timers = new Timer * [list.length()];

			while (num_timers < list.length ()) {
				e = list.item(num_timers).toElement ();

				timers[num_timers] = new Timer (&e);		
				event_map.insert (timers[num_timers]->ident, timers[num_timers]);
	
	        		num_timers++;
			}				
		}
	}


// now resolve the child-events in all events and timers
	for (unsigned int i=0; i < num_events; i++) {
		events[i]->resolve_now ();
	}
	for (unsigned int i=0; i < num_timers; i++) {
		timers[i]->resolve_now ();
	}

// read areas
	e = cdp->get_element ("hot_areas", xdocElement, Info);	// get all the hot-areas

	if (!(e.isNull())) {		// found a hot_areas-section?
		list = cdp->get_node_list ("area", e, 1, -1, 1, Warn);

		if (list.length ()) {
			areas = new HotArea * [list.length()];

			while (num_areas < list.length ()) {
				e = list.item(num_areas).toElement ();

				areas[num_areas] = new HotArea (&e);		
	
        			num_areas++;
			}
		}
	}


// END: events, states, areas

	e = cdp->get_element("platforms", xdocElement, Critical);	// get platforms section
	list = cdp->get_node_list("data", e, 2, -1, 2, Critical);	// get all platform data ( at least two platforms! )

	exp_fuel_platform = -1;
	if (list.length ()) {
		platform = new Platform * [list.length()];

		while (num_platforms < list.length ()) {		// get all data for this platform
			e = list.item(num_platforms).toElement();

			platform[num_platforms] = new Platform (&e, cdp->draw_highlight() && cdp->for_real ());		

			if (platform[num_platforms]->fuel_step) {
				if (exp_fuel_platform < 0) {
					exp_fuel_platform = num_platforms;
				} else if (platform[exp_fuel_platform]->fuel_price < platform[num_platforms]->fuel_price) {
					exp_fuel_platform = num_platforms;
				}
			}

			if (platform[num_platforms]->object_promised) {
				promise_pointer = new Promise;
				promise_pointer->ident = platform[num_platforms]->object_ident;
				promise_pointer->plat_point = platform[num_platforms];
				promise_pointer->type = 1;
				promises.append (promise_pointer);
			}
      	
			num_platforms++;
		}

		if (exp_fuel_platform < 0) {
			cdp->debug_msg(i18n("No fuel-platform specified. Assuming one on Platform 1"), Warn);
			platform[0]->fuel_step = 1;
			platform[0]->fuel_price = 0.1;
			exp_fuel_platform = 0;
		}
	}

	e = cdp->get_element("teleporters", xdocElement, Info);	// now get the Teleporters

	if (!(e.isNull())) {		// did we really find a "teleporters"-element?
		list = cdp->get_node_list("data", e, 2, -1, 2, Serious);	// get all teleporter data ( at least two if any! )

		if (list.length ()) {
			teleporters = new Teleporter* [list.length()];

			while (num_teleporters < list.length ()) {	// get all data for this teleporter
				e = list.item(num_teleporters).toElement();

				teleporters[num_teleporters] = new Teleporter (&e, list.length ());

				if (teleporters[num_teleporters]->object_promised) {
					promise_pointer = new Promise;
					promise_pointer->ident = teleporters[num_teleporters]->object_ident;
					promise_pointer->tp_point = teleporters[num_teleporters];
					promise_pointer->type = 0;
					promises.append (promise_pointer);
				}

				num_teleporters++;
			}
		}
	}


	e = cdp->get_element("objects", xdocElement, Info);	// now get the MovingObjects

	if (!(e.isNull())) {		// did we really find an objects-section?

		list = cdp->get_node_list("object", e, 1, -1, 1, Warn);	// get each single object in the section

		if (list.length ()) {
			objects = new MovingObject * [list.length()];

			while (num_objects < list.length ()) {				// create this object
				e = list.item (num_objects).toElement ();

				objects[num_objects] = new MovingObject ( &e );

				if (objects[num_objects]->has_ident) {
					promise_pointer = new Promise;
					promise_pointer->ident = objects[num_objects]->ident;
					promise_pointer->obj_point = objects[num_objects];
					definitions.append (promise_pointer);
				}

				num_objects++;
			}
		}
	}


	e = cdp->get_element("gravity", xdocElement, Serious);	// now get the gravity-section
	n_base_gravity = cdp->get_float_attribute("base_gravity", e, -5, 5, 0.01, Warn);
	n_horiz_gravity = cdp->get_float_attribute("horiz_gravity", e, -5, 5, 0, Warn);

	list = cdp->get_node_list("data", e, 0, -1, 0, Info);	// now get the Centers of Gravity

	if (list.length ()) {
		grav_center = new CenterOfGravity * [list.length()];

		while (num_grav_centers < list.length ()) {		// get all data for this center
			e = list.item(num_grav_centers).toElement();

			grav_center[num_grav_centers] = new CenterOfGravity (&e);

			if (grav_center[num_grav_centers]->object_promised) {
				promise_pointer = new Promise;
				promise_pointer->ident = grav_center[num_grav_centers]->object_ident;
				promise_pointer->cog_point = grav_center[num_grav_centers];
				promise_pointer->type = 2;
				promises.append (promise_pointer);
			}

			num_grav_centers++;
		}
	}

	e = cdp->get_element("tips", xdocElement, Warn);	// find the tips-tag
	min_base_tip = cdp->get_float_attribute("min", e, 0, 2000, 10, Warn);
	max_base_tip = cdp->get_float_attribute("max", e, 0, 2000, 30, Warn);

	e = cdp->get_element("fare", xdocElement, Warn);	// find the fare-tag
	base_fare = cdp->get_float_attribute("base", e, 0, 200, 4.5, Warn);
	fare_step = cdp->get_float_attribute("step", e, 0, 1, 0.01, Warn);

	e = cdp->get_element("initial_position", xdocElement, Serious);	// find the initial_position-tag
	initial_x = cdp->get_int_attribute("x", e, 0, 3000, 100, Serious);
	initial_y = cdp->get_int_attribute("y", e, 0, 3000, 100, Serious);

	setSequence(image);			// need this (width, height) before we can take care of oversized levels

	e = cdp->get_element("oversize", xdocElement, Info);
	if (!(e.isNull())) {
		visible_width = cdp->get_int_attribute("visible_width", e, 200, width(), width(), Warn);
		visible_height = cdp->get_int_attribute("visible_height", e, 200, height(), height(), Warn);
		max_offset_x = width() - visible_width;
		max_offset_y = height() - visible_height;
		initial_offset_x = cdp->get_int_attribute("initial_offset_x", e, 0, max_offset_x, 0, Info);
		initial_offset_y = cdp->get_int_attribute("initial_offset_y", e, 0, max_offset_y, 0, Info);
		margin = cdp->get_int_attribute("margin", e, 0, visible_width / 2, 50, Info);
	} else {					// this else of course has the same effect as the if, when the tag is not supplied.
		visible_width = width();	// it's only separate to suppress error-messages if the (optional) tag is not supplied.
		visible_height = height();
		max_offset_x = max_offset_y = initial_offset_x = initial_offset_y = margin = 0;
	}

// Finally, check whether all promises have been kept / assign promised objects
	
	promise_pointer = promises.first ();

	while (promise_pointer) {
		definitions.first();
		bool found = false;
		while (definitions.current() && !found) {
			if (definitions.current()->ident == promise_pointer->ident) {
				switch (promise_pointer->type) {
					case 0:	if (definitions.current ()->obj_point->teleporter) {
								found = true;
								promise_pointer->tp_point->promise_fulfilled (definitions.current()->obj_point);
								definitions.current ()->obj_point->teleporter_p = promise_pointer->tp_point;
								definitions.remove ();
						        }
							break;
					case 1:	if (definitions.current ()->obj_point->platform) {
								found = true;
								promise_pointer->plat_point->promise_fulfilled (definitions.current()->obj_point);
								definitions.current ()->obj_point->platform_p = promise_pointer->plat_point;
								definitions.remove ();
						        }
							break;
					case 2:	if (definitions.current ()->obj_point->center) {
								found = true;
								promise_pointer->cog_point->promise_fulfilled (definitions.current()->obj_point);
								definitions.remove ();
							}
							break;
					default:	break;
				}
			}			
			definitions.next();
		}
       	
		if (found) {
			promises.remove ();
			promise_pointer = promises.current();
		} else {
			QString type_str;
			switch (promise_pointer->type) {
				case 0:	type_str = "Teleporter";
						break;
				case 1:	type_str = "Platform";
						break;
				case 2:	type_str = "Center-of-Gravity";
						break;
				default:	type_str = bad_string_value;
						break;
			}
			cdp->debug_msg ("Promised object of type '" + type_str + "', named '" + promise_pointer->ident +
								"' was not defined.", Critical);
			promises.next ();
		}
	}	
	
	promise_pointer = definitions.first ();
	while (promise_pointer) {
		cdp->debug_msg ("Object named '" + promise_pointer->ident + "' was defined but never promised.", Warn);
		promise_pointer = definitions.next ();
	}

	state_map.clear ();				// do not call resolve_state () after this point!
	event_map.clear ();				// do not call resolve_event () after this point!

	cdp->debug_msg ("Finished loading level", Info);

}

/** No descriptions */
void Level::config_changed(){
	base_gravity = n_base_gravity * 2000.0 / ((float) cdp->cps () * (float) cdp->fps ());
	horiz_gravity = n_horiz_gravity * 2000.0 / ((float) cdp->cps () * (float) cdp->fps ());

	for (unsigned int i = 0; i < num_grav_centers; i++) {
		grav_center[i]->config_changed ();            	// Centers of gravity need to recalculate their "constants"
	}
}

/** Returns a pointer to the state identified with ident. Must never be called while not within read_level () ! */
State * Level::resolve_state (QString & ident){
	if (state_map.find (ident) == state_map.end ()) {
		cdp->debug_msg ( i18n ( "State '") + ident + i18n ( "' referenced but undeclared!" ), Critical);
		return 0;
	} else {
		return (state_map[ident]);
	}
}

/** Returns a pointer to the event identified with ident. Must never be called while not within read_level () ! */
Event * Level::resolve_event (QString & ident) {
	if (event_map.find (ident) == event_map.end ()) {
		cdp->debug_msg ( i18n ( "Event '") + ident + i18n ( "' referenced but undeclared!" ), Critical);
		return 0;
	} else {
		return (event_map[ident]);
	}
}